/* Main engine controll module, generate and make message map,
- control camera parametrs with keyboard.
- make collision detection.
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "texture.h"

#include "file.h"

#include "filerem.h"

#include "filepcx.h"

#include "lighttab.h"

#include "draw.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "socket.h"

#include "svsocket.h"

#include "nodever.h"

#include "map.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeint.h"

#include "nodetex.h"

#include "nodepor.h"

#include "nodeobj.h"

#include "nodever.h"

#include "filetxe.h"

#include "scenery.h"

#include "render.h"

#include "fly.h"



/* constructor - init empty
**********************************************************/
CFly::CFly ()
{
   mDestroy ();
}



/* constructor - init fly system
**********************************************************/
CFly::CFly ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   mCreate ( oScen, nResX, nResY, oFrBuff );
}



/* destructor
**********************************************************/
CFly::~CFly ()
{
   mDestroy ();
}



/* Init fly system.
oScen define the scenery to render
nResX define the X resolution to frame buffer
nResY define the Y resolution to frame buffer
oFrBuffer define the frame buffer
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CFly::mCreate ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   mDestroy ();

   m_poScen = &oScen; // define scenery

   if ( m_chSocketUse == 'c' )
      if ( ! m_oCliSoc.mCreate () )
         ABORT ( MSG_ERROR_SOCKET_CREATION, __FILE__, __LINE__ )
   
   return m_oRender.mCreate ( oScen, nResX, nResY, oFrBuff );
}



/* Destroy fly system.
**********************************************************/
void CFly::mDestroy ()
{
   m_poScen = NULL;
   m_chSocketUse = 'n';
   m_bConnected = FALSE;

   m_oRender.mDestroy ();
}



/* Create images, use keyboard control, make collision detection
return TRUE - ok / FALSE - error or end loop
**********************************************************/
BOOL CFly::mRunFrame ()
{
   // REVIEW (remover teste: overhead)
   // verify if not have scenery
   if ( m_poScen == NULL )
      return FALSE;

   // get the current camera from scenery

   // REVIEW (no caso ideal com portais nos objetos isso tem que mudar)
   // camera is inside this object
   CNodeObject
      *poParentObj = (CNodeObject*) m_poScen->m_oRootInstances.m_oInstances.mRetValue ( m_poScen->m_dwInsideObjID );

   // camera is inside this sector
   CNodeSector
      *poParentSec = (CNodeSector*) poParentObj->m_oSectors.mRetValue ( m_poScen->m_dwInsideSecID );

   // get the object that is the camera
   CNodeObject
      *poObj = (CNodeObject*) poParentSec->m_oInstances.mRetValue ( m_poScen->m_dwCameraID );

   // verify if no camera available
   if ( poObj == NULL )
      ABORT ( MSG_ERROR_NO_CAMERA, __FILE__, __LINE__ )

   // get the camera of the object
   CCameraSystem
      *poCamSys = (CCameraSystem*) &poObj->m_oRefSys;

   // verify if no camera available
   if ( poCamSys == NULL )
      ABORT ( MSG_ERROR_NO_CAMERA, __FILE__, __LINE__ )

   // define rotation angles (between 0 and 255)
   BYTE
      yAX = 0,
      yAY = 0,
      yAZ = 0;
   
   // move direction
   tagfXYZ
      stDirection;
   stDirection.fX = 0;
   stDirection.fY = 0;
   stDirection.fZ = 0;

   // keyboard input

   // verify keyboard for Linux target
   #ifdef TRG_LINUX   
   switch ( vga_getkey () ) {
      case 27:   // ESC
         // special key (91) or exit?
         if ( vga_getkey () != 91 )
            return FALSE;
         else
            switch ( vga_getkey () ) {
               case 'A':      // key UP
                  stDirection.fX += poCamSys->m_stJ.fX * poCamSys->m_fSpeed;
                  stDirection.fY += poCamSys->m_stJ.fY * poCamSys->m_fSpeed;
                  stDirection.fZ += poCamSys->m_stJ.fZ * poCamSys->m_fSpeed;
                  break;
               case 'B':      // key DOWN
                  stDirection.fX -= poCamSys->m_stJ.fX * poCamSys->m_fSpeed;
                  stDirection.fY -= poCamSys->m_stJ.fY * poCamSys->m_fSpeed;
                  stDirection.fZ -= poCamSys->m_stJ.fZ * poCamSys->m_fSpeed;
                  break;
               case 'C':      // key RIGHT
                  stDirection.fX += poCamSys->m_stI.fX * poCamSys->m_fSpeed;
                  stDirection.fY += poCamSys->m_stI.fY * poCamSys->m_fSpeed;
                  stDirection.fZ += poCamSys->m_stI.fZ * poCamSys->m_fSpeed;
                  break;
               case 'D':      // key LEFT
                  stDirection.fX -= poCamSys->m_stI.fX * poCamSys->m_fSpeed;
                  stDirection.fY -= poCamSys->m_stI.fY * poCamSys->m_fSpeed;
                  stDirection.fZ -= poCamSys->m_stI.fZ * poCamSys->m_fSpeed;
                  break;
            }
         break;

      // forward
      case 'A':
      case 'a':
         stDirection.fX += poCamSys->m_stK.fX * poCamSys->m_fSpeed;
         stDirection.fY += poCamSys->m_stK.fY * poCamSys->m_fSpeed;
         stDirection.fZ += poCamSys->m_stK.fZ * poCamSys->m_fSpeed;
         break;

      // backward
      case 'Z':
      case 'z':
         stDirection.fX -= poCamSys->m_stK.fX * poCamSys->m_fSpeed;
         stDirection.fY -= poCamSys->m_stK.fY * poCamSys->m_fSpeed;
         stDirection.fZ -= poCamSys->m_stK.fZ * poCamSys->m_fSpeed;
         break;

      // ROTATE
      case '8':       // key UP
         yAX -= 1;
         break;
      case '2':       // key DOWN
         yAX += 1;
         break;
      case '6':
         yAY -= 1;
         break;
      case '4':
         yAY += 1;
         break;
      case '7':
         yAZ += 1;
         break;
      case '9':
         yAZ -= 1;
         break;
   }
   #endif

   // verify keyboard for DJGPP target
   #ifdef TRG_DJGPP
   // test exit
   if ( m_oKb.mTestKeyPress ( K_ESC ) )
      return FALSE;

   // test ALT
   if ( m_oKb.mTestKeyPress ( K_ALT ) || m_oKb.mTestKeyPress ( K_ALT | SPECIAL ) ) {
      // move
      if ( m_oKb.mTestKeyPress ( K_UP ) || m_oKb.mTestKeyPress ( K_UP | SPECIAL ) ) {
         stDirection.fX += poCamSys->m_stJ.fX * poCamSys->m_fSpeed;
         stDirection.fY += poCamSys->m_stJ.fY * poCamSys->m_fSpeed;
         stDirection.fZ += poCamSys->m_stJ.fZ * poCamSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_DOWN ) || m_oKb.mTestKeyPress ( K_DOWN | SPECIAL ) ) {
         stDirection.fX -= poCamSys->m_stJ.fX * poCamSys->m_fSpeed;
         stDirection.fY -= poCamSys->m_stJ.fY * poCamSys->m_fSpeed;
         stDirection.fZ -= poCamSys->m_stJ.fZ * poCamSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_RIGHT ) || m_oKb.mTestKeyPress ( K_RIGHT | SPECIAL ) ) {
         stDirection.fX += poCamSys->m_stI.fX * poCamSys->m_fSpeed;
         stDirection.fY += poCamSys->m_stI.fY * poCamSys->m_fSpeed;
         stDirection.fZ += poCamSys->m_stI.fZ * poCamSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_LEFT ) || m_oKb.mTestKeyPress ( K_LEFT | SPECIAL ) ) {
         stDirection.fX -= poCamSys->m_stI.fX * poCamSys->m_fSpeed;
         stDirection.fY -= poCamSys->m_stI.fY * poCamSys->m_fSpeed;
         stDirection.fZ -= poCamSys->m_stI.fZ * poCamSys->m_fSpeed;
      }
   }
   else {
      // rotations
      if ( m_oKb.mTestKeyPress ( K_UP ) || m_oKb.mTestKeyPress ( K_UP | SPECIAL ) )
         yAX -= 1;
      if ( m_oKb.mTestKeyPress ( K_DOWN ) || m_oKb.mTestKeyPress ( K_DOWN | SPECIAL ) )
         yAX += 1;
      if ( m_oKb.mTestKeyPress ( K_RIGHT ) || m_oKb.mTestKeyPress ( K_RIGHT | SPECIAL ) )
         yAY -= 1;
      if ( m_oKb.mTestKeyPress ( K_LEFT ) || m_oKb.mTestKeyPress ( K_LEFT | SPECIAL ) )
         yAY += 1;
   }
   if ( m_oKb.mTestKeyPress ( K_HOME ) )
      yAZ += 1;
   if ( m_oKb.mTestKeyPress ( K_PG_UP ) )
      yAZ -= 1;
   if ( m_oKb.mTestKeyPress ( K_A ) ) {
      stDirection.fX += poCamSys->m_stK.fX * poCamSys->m_fSpeed;
      stDirection.fY += poCamSys->m_stK.fY * poCamSys->m_fSpeed;
      stDirection.fZ += poCamSys->m_stK.fZ * poCamSys->m_fSpeed;
   }
   if ( m_oKb.mTestKeyPress ( K_Z ) ) {
      stDirection.fX -= poCamSys->m_stK.fX * poCamSys->m_fSpeed;
      stDirection.fY -= poCamSys->m_stK.fY * poCamSys->m_fSpeed;
      stDirection.fZ -= poCamSys->m_stK.fZ * poCamSys->m_fSpeed;
   }
   #endif

   // update camera

   // detect collision
   mDetectCollision ( *poObj, m_poScen->m_dwCameraID, *poParentSec, stDirection );
   // change position
   poCamSys->m_stPos.fX += stDirection.fX;
   poCamSys->m_stPos.fY += stDirection.fY;
   poCamSys->m_stPos.fZ += stDirection.fZ;
   // rotate
   poCamSys->mRotateTransBase ( yAX, yAY, yAZ );

   // socket communication
   
   // verify if is to make socket communication
   if ( m_chSocketUse != 'n' )

      // if is the client
      if ( m_chSocketUse == 'c' ) {
         
         // make socket communication and update
         // verify if is connected
         if ( m_bConnected ) {
    
            // send my position
            // invert the camera
            m_oSocCam.mSetObj ( *poCamSys );
            m_poScen->mInvertCamera ( m_oSocCam );
            // copy the camera definition to a buffer to be send
            float
               afCom [ 13 ];
            afCom [ 0 ] = m_oSocCam.m_stPos.fX;
            afCom [ 1 ] = m_oSocCam.m_stPos.fY;
            afCom [ 2 ] = m_oSocCam.m_stPos.fZ;
            afCom [ 3 ] = m_oSocCam.m_stI.fX;
            afCom [ 4 ] = m_oSocCam.m_stI.fY;
            afCom [ 5 ] = m_oSocCam.m_stI.fZ;
            afCom [ 6 ] = m_oSocCam.m_stJ.fX;
            afCom [ 7 ] = m_oSocCam.m_stJ.fY;
            afCom [ 8 ] = m_oSocCam.m_stJ.fZ;
            afCom [ 9 ] = m_oSocCam.m_stK.fX;
            afCom [10 ] = m_oSocCam.m_stK.fY;
            afCom [11 ] = m_oSocCam.m_stK.fZ;
            afCom [12 ] = (float) m_poScen->m_dwCameraID; // send my ID
            // send
            m_oCliSoc.mSend ( (void*) afCom, sizeof afCom );

            // read from socket the other camera position
            INT
               nCode;
            // if read some data
            if ( m_oCliSoc.mReceive ( (void*) afCom, sizeof afCom, nCode ) ) {

               // verify if connection ended or error
               if ( nCode <= 0 ) {
                  m_oCliSoc.mDestroy ();
                  m_bConnected = FALSE;
               }
               else { // if connection ok

                  // change the other camera position
                  DWORD
                     dwOtherID = (DWORD) afCom [ 12 ];
                  // search the object
                  CNodeObject
                     *poOther = (CNodeObject*) m_poScen->m_oRootInstances.mSearchObject ( dwOtherID );
                  // verify if object exist  // REVIEW
                  if ( poOther != NULL ) {
                     // copy the given definition
                     poOther->m_oRefSys.m_stPos.fX = afCom [ 0 ];
                     poOther->m_oRefSys.m_stPos.fY = afCom [ 1 ];
                     poOther->m_oRefSys.m_stPos.fZ = afCom [ 2 ];
                     // REVIEW (no servidor tambem!!!)
                     /*poOther->m_oRefSys.m_stI.fX = afCom [ 3 ];
                     poOther->m_oRefSys.m_stI.fY = afCom [ 4 ];
                     poOther->m_oRefSys.m_stI.fZ = afCom [ 5 ];
                     poOther->m_oRefSys.m_stJ.fX = afCom [ 6 ];
                     poOther->m_oRefSys.m_stJ.fY = afCom [ 7 ];
                     poOther->m_oRefSys.m_stJ.fZ = afCom [ 8 ];
                     poOther->m_oRefSys.m_stK.fX = afCom [ 9 ];
                     poOther->m_oRefSys.m_stK.fY = afCom [10 ];
                     poOther->m_oRefSys.m_stK.fZ = afCom [11 ];*/
                     poOther->m_oRefSys.mComputeFixPointBase ();
                  }
               }  // END IF connection ok
            }
         }
      }
      else {   // if this is the server
         
         // make socket communication and update
         // verify if is connected -> make communication
         if ( m_bConnected ) {
            
            // send my position
            // invert the camera
            m_oSocCam.mSetObj ( *poCamSys );
            m_poScen->mInvertCamera ( m_oSocCam );
            // copy the camera definition to a buffer to be send
            float
               afCom [ 13 ];
            afCom [ 0 ] = m_oSocCam.m_stPos.fX;
            afCom [ 1 ] = m_oSocCam.m_stPos.fY;
            afCom [ 2 ] = m_oSocCam.m_stPos.fZ;
            afCom [ 3 ] = m_oSocCam.m_stI.fX;
            afCom [ 4 ] = m_oSocCam.m_stI.fY;
            afCom [ 5 ] = m_oSocCam.m_stI.fZ;
            afCom [ 6 ] = m_oSocCam.m_stJ.fX;
            afCom [ 7 ] = m_oSocCam.m_stJ.fY;
            afCom [ 8 ] = m_oSocCam.m_stJ.fZ;
            afCom [ 9 ] = m_oSocCam.m_stK.fX;
            afCom [10 ] = m_oSocCam.m_stK.fY;
            afCom [11 ] = m_oSocCam.m_stK.fZ;
            afCom [12 ] = (float) m_poScen->m_dwCameraID; // send my ID
            // send
            m_oSrvSoc.mSend ( (void*) afCom, sizeof afCom );

            // read from socket the other camera position
            INT
               nCode;

            // if read some data
            if ( m_oSrvSoc.mReceive ( (void*) afCom, sizeof afCom, nCode ) ) {

               // verify if connection ended or error
               if ( nCode <= 0 ) {
                  m_oSrvSoc.mDestroy ();
                  m_bConnected = FALSE;
               }
               else { // if connection ok

                  // change the other camera position
                  DWORD
                     dwOtherID = (DWORD) afCom [ 12 ];
                  // search the object
                  CNodeObject
                     *poOther = (CNodeObject*) m_poScen->m_oRootInstances.mSearchObject ( dwOtherID );
                  // verify if object exist  // REVIEW
                  if ( poOther != NULL ) {
                  // copy the given definition
                  poOther->m_oRefSys.m_stPos.fX = afCom [ 0 ];
                  poOther->m_oRefSys.m_stPos.fY = afCom [ 1 ];
                  poOther->m_oRefSys.m_stPos.fZ = afCom [ 2 ];
                  /*poOther->m_oRefSys.m_stI.fX = afCom [ 3 ];
                  poOther->m_oRefSys.m_stI.fY = afCom [ 4 ];
                  poOther->m_oRefSys.m_stI.fZ = afCom [ 5 ];
                  poOther->m_oRefSys.m_stJ.fX = afCom [ 6 ];
                  poOther->m_oRefSys.m_stJ.fY = afCom [ 7 ];
                  poOther->m_oRefSys.m_stJ.fZ = afCom [ 8 ];
                  poOther->m_oRefSys.m_stK.fX = afCom [ 9 ];
                  poOther->m_oRefSys.m_stK.fY = afCom [10 ];
                  poOther->m_oRefSys.m_stK.fZ = afCom [11 ];*/
                  poOther->m_oRefSys.mComputeFixPointBase ();
               }
            }
         }
      }   // END IF server
   }      // END IF use socket (!='n')

   // make image
   m_oRender.mRenderFrame ( m_poScen->m_dwInsideObjID, m_poScen->m_dwInsideSecID, poCamSys );

   return TRUE;
}


/* Make collision detection verifying the distance between de camera
center and the current sector polygons. Detect portal crossing and
change the current camera object/sector if needed.
oObj - the camera object to make test
dwObjID - the object ID
oParentObj and oParentSec - where the camera is
stDirection - is the direction vector where camera want move
**********************************************************/
void CFly::mDetectCollision ( CNodeObject& oObj, DWORD dwObjID,
                              CNodeSector& oParentSec, tagfXYZ& stDirection )
{
   // scan all polygons in sector verifying the distance
   // for each polygon in mesh
   oParentSec.m_oPolygons.mSelectStart ();
   // current polygon in test
   DWORD
      dwPolyNdx = 0; 
   while ( oParentSec.m_oPolygons.mRetExistSelection () ) {

      // the current polygon to test
      CNodePolygon
         *poPol = (CNodePolygon*) oParentSec.m_oPolygons.mRetSelected ();
      
      // the new camera position without test
      tagfXYZ
         stPosNew;
      stPosNew.fX = oObj.m_oRefSys.m_stPos.fX + stDirection.fX;
      stPosNew.fY = oObj.m_oRefSys.m_stPos.fY + stDirection.fY;
      stPosNew.fZ = oObj.m_oRefSys.m_stPos.fZ + stDirection.fZ;

      // compute the distance between plan and camera position
      float
         fDistance = - ( - poPol->m_stNormal.fX * stPosNew.fX +
                         - poPol->m_stNormal.fY * stPosNew.fY +
                         - poPol->m_stNormal.fZ * stPosNew.fZ +
                           poPol->m_fD );

      // if the polygon is not crossable and you are inside your plan
      // change the direction vector acording the polygon normal
      if ( poPol->m_yCrossable != 0 &&                    // crossable test
           fDistance <= oObj.m_oRefSys.m_fRadius ) {      // inside test

         // flag if point is inside (TRUE) or outside poly (FALSE)
         BOOL
            bPointInside = TRUE;
         // verify collision detection testing if point is inside poly
         if ( poPol->m_yCrossable == 2 ) {

            // need test if point is inside poly:
            // verify if the intersection point is inside polygon using
            // the vectors signals

            // number of vertexes in poly
            DWORD
               dwNumVerts = poPol->m_dwNumVertexes;

            // the intersection point between plan and sphere (projected center
            // of sphere to poly plan)
            tagXYZ
               stPoint;
            stPoint.nX = (INT) ( stPosNew.fX - poPol->m_stNormal.fX * fDistance );
            stPoint.nY = (INT) ( stPosNew.fY - poPol->m_stNormal.fY * fDistance );
            stPoint.nZ = (INT) ( stPosNew.fZ - poPol->m_stNormal.fZ * fDistance );

            // point poly vertex index array
            DWORD
               *pdwVertNdx = poPol->m_padwVertexes;
            // vertexes to work with         
            CNodeVertex
               *poVert1 = (CNodeVertex*) oParentSec.m_oVertexes.mRetNode ( *pdwVertNdx ),
               *poVert2 = NULL;
            // flag dot-product sign (if all have the same sign the point is inside poly)
            BOOL
               bFlagDot = TRUE;  // TRUE - positive, FALSE - negative
            // for each vertex in poly make the test (scan all poly segments)
            for ( DWORD dwcV = 0; dwcV < dwNumVerts; dwcV ++ ) {
            // -1 because you need the current vertex and the next to test

               // define next vertex
               if ( dwcV == dwNumVerts - 1 )  // if last index
                  pdwVertNdx = poPol->m_padwVertexes; // use first
               else     
                  pdwVertNdx ++;                            // next
               // point it
               poVert2 = (CNodeVertex*) oParentSec.m_oVertexes.mRetNode ( *pdwVertNdx );

               // define 2 vectors using the vertexes and the point
               tagXYZ
                  stVector1,
                  stVector2;
               stVector1.nX = poVert1->m_stVertex.nX - stPoint.nX;
               stVector1.nY = poVert1->m_stVertex.nY - stPoint.nY;
               stVector1.nZ = poVert1->m_stVertex.nZ - stPoint.nZ;
               stVector2.nX = poVert1->m_stVertex.nX - poVert2->m_stVertex.nX;
               stVector2.nY = poVert1->m_stVertex.nY - poVert2->m_stVertex.nY;
               stVector2.nZ = poVert1->m_stVertex.nZ - poVert2->m_stVertex.nZ;

               // REVIEW - does not work for all cases
               // compute dot-product
               INT
                  nDotProd = stVector1.nX * stVector2.nX +
                             stVector1.nY * stVector2.nY +
                             stVector1.nZ * stVector2.nZ;

               // verify if is the first dot calculated (only update signal flags)
               if ( dwcV == 0 && nDotProd < 0 )
                  bFlagDot = FALSE;
               else  // if not the first, make sign comparisons
                  if ( ( bFlagDot == TRUE && nDotProd < 0 ) ||
                       ( bFlagDot == FALSE && nDotProd >= 0 ) ) {
                     bPointInside = FALSE;
                     break;
                  }

               // update who is the first vertex
               poVert1 = poVert2;
            }
         }
            
         // verify collision detection:
         // if only need test plan distance (poPol->m_stPol.bCrossable==1)
         // if need test if intersection point is inside poly (bPointInside TRUE)
         if ( poPol->m_yCrossable == 1 || bPointInside ) {
            stDirection.fX += poPol->m_stNormal.fX * oObj.m_oRefSys.m_fSpeed;
            stDirection.fY += poPol->m_stNormal.fY * oObj.m_oRefSys.m_fSpeed;
            stDirection.fZ += poPol->m_stNormal.fZ * oObj.m_oRefSys.m_fSpeed;
         }
      }
      else
         // if the polygon is crossable, is a portal and you are behind it
         // make portal crossing detection: change current sector if distance
         // is negative
         if ( poPol->m_yCrossable == 0 &&         // crossable test
              poPol->m_bPortal == TRUE &&         // portal
              fDistance < 0 ) {                         // negative distance (behind)
            // the new object and sector where you are
            // define the object and sector IDs
            CNodePortal
               *poDestin = (CNodePortal*) oParentSec.m_oPortals.mRetValue ( dwPolyNdx );
            // move the instance to the sector
            m_poScen->mMoveInstance ( oObj, dwObjID, oParentSec,
                                      poDestin->m_stPortal.dwObjectID,
                                      poDestin->m_stPortal.dwSectorID );
            // no more object inside this mesh
            return;
         }

      // next polygon to test
      oParentSec.m_oPolygons.mSelectNext ();
      dwPolyNdx ++;
   }
}
